@@ -107,7 +107,7 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name, int
 				end--;
 
 			o_len = end - cp + 1;
-			if (strncmp("symlink", cp, o_len) == 0)
+			if (_strncmp("symlink", cp, o_len) == 0)
 				symlnk = 1;
 		}
 	}
@@ -135,6 +135,12 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name, int
 	if (options == NULL || *options == '\0')
 		options = "defaults";
 
+	if (!strcmp(what, fullpath)) {
+		debug(ap->logopt, MODPREFIX
+		     "cannot mount or symlink %s to itself", fullpath);
+		return 1;
+	}
+
 	if (!symlnk && bind_works) {
 		int status, existed = 1;
 
@@ -206,16 +212,16 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name, int
 		if (symlink(what, fullpath) && errno != EEXIST) {
 			error(ap->logopt,
 			      MODPREFIX
-			      "failed to create local mount %s -> %s",
+			      "failed to create symlink %s -> %s",
 			      fullpath, what);
-			if ((ap->flags & MOUNT_FLAG_GHOST) && !status)
+			if ((ap->flags & MOUNT_FLAG_GHOST) && !status) {
 				if (mkdir_path(fullpath, 0555) && errno != EEXIST) {
 					char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 					error(ap->logopt,
 					      MODPREFIX "mkdir_path %s failed: %s",
 					      fullpath, estr);
 				}
-			else {
+			} else {
 				if (ap->type == LKP_INDIRECT)
 					rmdir_path(ap, fullpath, ap->dev);
 			}
